{-# OPTIONS_GHC -Wall #-}
module Formula where

import qualified Data.Set as S 
import Data.List

data Formula
    = F
    | T
    | A String
    | Imp Formula Formula
    | And Formula Formula
    | Or Formula Formula
    | Not Formula
    | Coimp Formula Formula
    deriving (Show, Eq, Ord)

data Sequent = S (S.Set Formula) Formula
    deriving (Show, Eq, Ord)

prettyF :: Formula -> String
prettyF F           = "F"
prettyF T           = "T"
prettyF (A x)       = x
prettyF (Imp p q)   = "(" ++ prettyF p ++ " -> "  ++ prettyF q ++ ")"
prettyF (And p q)   = "(" ++ prettyF p ++ " /\\ " ++ prettyF q ++ ")"
prettyF (Or  p q)   = "(" ++ prettyF p ++ " \\/ " ++ prettyF q ++ ")"
prettyF (Not p)     = "~" ++ prettyF p
prettyF (Coimp p q) = "(" ++ prettyF p ++ " ->> " ++ prettyF q ++ ")"

prettyS :: Sequent -> String
prettyS (S pre p) = 
  (concat $ intersperse ", " $ map prettyF $ S.toList pre)
  ++ " |- " ++ prettyF p

-- | LaTeX variant
prettyLF :: Formula -> String
prettyLF F           = "\\bot "
prettyLF T           = "\\top "
prettyLF (A x)       = x -- "\\mathsf{" ++ x ++ "}" -- FIXME no escape...
prettyLF (Imp p q)   = "(" ++ prettyLF p ++ " \\imp "   ++ prettyLF q ++ ")"
prettyLF (And p q)   = "(" ++ prettyLF p ++ " \\land "  ++ prettyLF q ++ ")"
prettyLF (Or  p q)   = "(" ++ prettyLF p ++ " \\lor "   ++ prettyLF q ++ ")"
prettyLF (Not p)     = "\\lnot " ++ prettyLF p
prettyLF (Coimp p q) = "(" ++ prettyLF p ++ " \\coimp " ++ prettyLF q ++ ")"

prettyLS :: Sequent -> String
prettyLS (S pre p) = 
  (concat $ intersperse ", " $ map prettyLF $ S.toList pre)
  ++ " \\vdash " ++ prettyLF p

atomsF :: Formula -> S.Set String
atomsF F = S.empty
atomsF T = S.empty
atomsF (A x) = S.singleton x
atomsF (Imp p q) = atomsF p `S.union` atomsF q
atomsF (And p q) = atomsF p `S.union` atomsF q
atomsF (Or  p q) = atomsF p `S.union` atomsF q
atomsF (Not p) = atomsF p
atomsF (Coimp p q) = atomsF p `S.union` atomsF q

atoms :: S.Set Formula -> S.Set String
atoms = S.unions . S.toList . S.map atomsF

genFresh :: S.Set Formula -> String -> Formula
genFresh ps pref = A $ head $ filter (not . (`S.member` atoms ps))
                   [ pref ++ show i | i <- [0::Integer ..] ]
