{-# OPTIONS_GHC -Wall #-}
module Parser (formula) where

import Formula
import Control.Monad
import Text.ParserCombinators.Parsec

-- | Imp @=>@, Coimp @->@, And @/\@, Or @\/@
connective :: Parser (Formula -> Formula -> Formula)
connective =
    do op <- (try (string "->>") >> return Coimp) <|>
             (string "->" >> return Imp) <|>
             (string "/\\" >> return And) <|>
             (string "\\/" >> return Or)
       spaces
       return op

-- | @( p OP q)@
binary :: Parser Formula
binary =
    do char '(' >> spaces
       p <- form
       op <- connective
       q <- form
       char ')' >> spaces
       return (p `op` q)

-- | Not @~ p@
negation :: Parser Formula
negation =
    do char '~' >> spaces
       p <- form
       return (Not p)

-- | Letters allowed in prime names
primeLetter :: Parser Char
primeLetter = letter <|> digit <|> oneOf "'"

-- | @p@
prime :: Parser Formula
prime =
    do p <- many1 primeLetter
       spaces
       return (A p)

-- | False
false :: Parser Formula
false = try (char 'F' >> notFollowedBy primeLetter)
        >> spaces >> return F

-- | True
true :: Parser Formula
true = try (char 'T' >> notFollowedBy primeLetter)
        >> spaces >> return T

-- | Sub-formula
form :: Parser Formula
form = binary <|> negation <|> false <|> true <|> prime

-- | Top-level formula
formula :: Parser Formula
formula =
    do spaces
       f <- form
       eof
       return f