{-# OPTIONS_GHC -Wall #-}
module Main where

import Control.Monad (when, unless)
-- import Data.List (elem)
import Data.Maybe
import System.Environment
import System.Exit
import System.Console.GetOpt
import qualified Text.ParserCombinators.Parsec as P

import Formula
import Proof
import Parser

data Flag = Verbosity Int | Help | Vers | Err String
          deriving Eq

readVerb :: String -> Flag
readVerb s = case reads s of
               [(n,"")] -> Verbosity n
               _        -> Err "Invalid verbosity level"

optDescr :: [OptDescr Flag]
optDescr = [ Option "" ["verbosity"] (ReqArg readVerb "n") "Verbosity level"
           , Option "" ["help"]      (NoArg Help) "Display help"
           , Option "" ["version"]   (NoArg Vers) "Display version"
           ]

die :: [String] -> IO ()
die e =
    do putStr $ "Pcl: " ++ unlines e
       exitFailure

parseVerb :: [Flag] -> Int
parseVerb [] = 1
parseVerb (Verbosity n : _ ) = n
parseVerb (_           : xs) = parseVerb xs

isErr :: Flag -> Maybe String
isErr (Err e) = Just e
isErr _       = Nothing

main :: IO ()
main = do
  -- case testChain 9 of
  --  Result _ -> putStrLn "Result ***"
  --  x        -> putStrLn $ prettyR x
  -- putStrLn $ prettyR test5
  -- putStrLn $ prettyR test4
  args <- getArgs
  let (opt, fs, err) = getOpt Permute optDescr args
  let err2 = err ++ (catMaybes $ map isErr opt)
  unless (null err2) $ die err2
  when (Help `elem` opt) $ die [usageInfo "Pcl [options] formula" optDescr]
  when (Vers `elem` opt) $ die ["version 0.0.1.0"]
  let verb = parseVerb opt
  case fs of
    [f] -> prove verb f
    _   -> die ["A single formula must be specified."]

{-
  case args of
    [v,a] -> do verb <- parseVerbosity v
                prove verb a
    [a]   -> prove 1 a
    _     -> do putStrLn "usage: Pcl [--verbosity=n] <formula>"
                exitFailure
-}

prove :: Int -> String -> IO ()
prove verb a = do
  case P.parse formula "" a of
    Left e  -> do putStrLn $ "Pcl: parse error: " ++ show e
                  exitFailure
    Right p -> do when (verb>=1) $
                       putStrLn $ "Formula: " ++ prettyF p
                  res <- decideWithProgress (progressVerb verb) p
                  when (verb>=1) $
                       putStrLn $ prettyRVerb verb res
                  case res of
                    Result _ -> exitWith ExitSuccess
                    Timeout  -> error "decide returned timeout!"
                    NoResult -> exitFailure

{-
parseVerbosity :: String -> IO Int
parseVerbosity s =
    let verbOpt = "--verbosity="
        in if verbOpt `isPrefixOf` s
           then return $ read $ drop (length verbOpt) s
           else do putStrLn "Coimpl: cannot parse verbosity"
                   exitFailure
-}

prettyRVerb :: Int -> Result Proof -> String
prettyRVerb n (Result _) | n<3 = "Result"
prettyRVerb 3 x          = prettyR x
prettyRVerb _ x          = prettyLR x

progressVerb :: Int -> (Int -> IO ())
progressVerb n | n < 2     = const (return ())
               | otherwise = \i -> print i

-- * tests

test0 :: Result Proof
test0 = decide (Imp (Coimp p p) p)
        where p = A "p"

test1 :: Result Proof
test1 = decide (Imp (Coimp p q) (Imp (Coimp q p) p))
        where p = A "p"
              q = A "q"

test2 :: Result Proof
test2 = decide (Imp (Coimp q q) (Imp (Coimp q p) p))
        where p = A "p"
              q = A "q"

test3 :: Result Proof
test3 = decide (Imp (Coimp p q) q)
        where p = A "p"
              q = A "q"

test4 :: Result Proof
test4 = decide (Imp (Coimp p0 p1)
                (Imp (Coimp p1 p2)
                 (Imp (Coimp p2 p3)
                  (Imp (Coimp p3 p0)
                   (p0)))))
        where p0 = A "p0"
              p1 = A "p1"
              p2 = A "p2"
              p3 = A "p3"

test5 :: Result Proof
test5 = decide (Imp (Coimp p q)
            (Imp p q))
        where p = A "p"
              q = A "q"

testFix, testPrePost, testZero :: Result Proof
testFix = decide (Imp (Coimp p p) p)
    where p = A "p"
testPrePost = decide (Imp (Imp p' p)
                      (Imp (Coimp p q)
                       (Imp (Imp q q')
                        (Coimp p' q'))))
    where p = A "p"
          p'= A "p'"
          q = A "q"
          q'= A "q'"
testZero = decide (Coimp (Not F) (Not F))


testChain :: Int -> Result Proof
testChain n = decide $ (Imp (go 0) (p 0))
  where go x | x /= n = (And (Coimp (p x) (p (succ x)))
                             (go (succ x)))
             | otherwise = (Coimp (p x) (p 0))
        p :: Int -> Formula
        p x = A ('p' : show x)
